package com.linbit.linstor.layer.storage.spdk;

import com.linbit.ImplementationError;
import com.linbit.SizeConv;
import com.linbit.SizeConv.SizeUnit;
import com.linbit.extproc.ExtCmdFactoryStlt;
import com.linbit.linstor.PriorityProps;
import com.linbit.linstor.api.ApiConsts;
import com.linbit.linstor.api.SpaceInfo;
import com.linbit.linstor.backupshipping.BackupShippingMgr;
import com.linbit.linstor.clone.CloneService;
import com.linbit.linstor.core.StltConfigAccessor;
import com.linbit.linstor.core.apicallhandler.StltExtToolsChecker;
import com.linbit.linstor.core.devmgr.pojos.LocalNodePropsChangePojo;
import com.linbit.linstor.core.identifier.ResourceName;
import com.linbit.linstor.core.identifier.VolumeNumber;
import com.linbit.linstor.core.objects.Resource;
import com.linbit.linstor.core.objects.ResourceDefinition;
import com.linbit.linstor.core.objects.ResourceGroup;
import com.linbit.linstor.core.objects.Snapshot;
import com.linbit.linstor.core.objects.SnapshotVolume;
import com.linbit.linstor.core.objects.StorPool;
import com.linbit.linstor.core.objects.Volume;
import com.linbit.linstor.core.objects.VolumeDefinition;
import com.linbit.linstor.dbdrivers.DatabaseException;
import com.linbit.linstor.layer.DeviceLayer.NotificationListener;
import com.linbit.linstor.layer.DeviceLayerUtils;
import com.linbit.linstor.layer.storage.AbsStorageProvider;
import com.linbit.linstor.layer.storage.WipeHandler;
import com.linbit.linstor.layer.storage.spdk.utils.SpdkConfigReader;
import com.linbit.linstor.layer.storage.spdk.utils.SpdkUtils;
import com.linbit.linstor.layer.storage.spdk.utils.SpdkUtils.LvsInfo;
import com.linbit.linstor.logging.ErrorReporter;
import com.linbit.linstor.propscon.InvalidKeyException;
import com.linbit.linstor.propscon.Props;
import com.linbit.linstor.security.AccessContext;
import com.linbit.linstor.security.AccessDeniedException;
import com.linbit.linstor.snapshotshipping.SnapshotShippingService;
import com.linbit.linstor.storage.StorageConstants;
import com.linbit.linstor.storage.StorageException;
import com.linbit.linstor.storage.data.provider.spdk.SpdkData;
import com.linbit.linstor.storage.interfaces.categories.resource.VlmProviderObject.Size;
import com.linbit.linstor.storage.kinds.DeviceProviderKind;
import com.linbit.linstor.transaction.manager.TransactionMgr;

import static com.linbit.linstor.layer.storage.spdk.utils.SpdkUtils.SPDK_PATH_PREFIX;

import javax.inject.Provider;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class AbsSpdkProvider<T> extends AbsStorageProvider<LvsInfo, SpdkData<Resource>, SpdkData<Snapshot>>
{
    private static final int TOLERANCE_FACTOR = 3;
    // FIXME: FORMAT should be private, only made public for LayeredSnapshotHelper
    public static final String FORMAT_RSC_TO_SPDK_ID = "%s%s_%05d";
    public static final String FORMAT_SNAP_TO_SPDK_ID = FORMAT_RSC_TO_SPDK_ID + "_%s";
    private static final String FORMAT_SPDK_ID_WIPE_IN_PROGRESS = "%s-linstor_wiping_in_progress";
    private static final String SPDK_FORMAT_DEV_PATH = SPDK_PATH_PREFIX + "%s/%s";

    private static final String DFLT_LVCREATE_TYPE = "linear";
    protected final SpdkCommands<T> spdkCommands;

    protected AbsSpdkProvider(
        ErrorReporter errorReporter,
        ExtCmdFactoryStlt extCmdFactory,
        AccessContext storDriverAccCtx,
        StltConfigAccessor stltConfigAccessor,
        WipeHandler wipeHandler,
        Provider<NotificationListener> notificationListenerProvider,
        Provider<TransactionMgr> transMgrProvider,
        String subTypeDescr,
        DeviceProviderKind subTypeKind,
        SnapshotShippingService snapShipMrgRef,
        StltExtToolsChecker extToolsCheckerRef,
        SpdkCommands<T> spdkCommandsRef,
        CloneService cloneServiceRef,
        BackupShippingMgr backupShipMgrRef
    )
    {
        super(
            errorReporter,
            extCmdFactory,
            storDriverAccCtx,
            stltConfigAccessor,
            wipeHandler,
            notificationListenerProvider,
            transMgrProvider,
            subTypeDescr,
            subTypeKind,
            snapShipMrgRef,
            extToolsCheckerRef,
            cloneServiceRef,
            backupShipMgrRef
        );
        spdkCommands = spdkCommandsRef;
        isDevPathExpectedToBeNull = true;
    }

    @Override
    protected void updateStates(List<SpdkData<Resource>> vlmDataList, List<SpdkData<Snapshot>> snapshots)
        throws StorageException, AccessDeniedException, DatabaseException
    {
        final Map<String, Long> extentSizes = SpdkUtils.getExtentSize(
            spdkCommands,
            getAffectedVolumeGroups(vlmDataList, snapshots)
        );

        List<SpdkData<?>> combinedList = new ArrayList<>();
        combinedList.addAll(vlmDataList);
        combinedList.addAll(snapshots);

        for (SpdkData<?> vlmData : combinedList)
        {

            final LvsInfo info = infoListCache.get(getFullQualifiedIdentifier(vlmData));
            updateInfo(vlmData, info);

            if (info != null)
            {
                final long expectedSize = vlmData.getExpectedSize();
                final long actualSize = info.size;
                if (actualSize != expectedSize)
                {
                    if (actualSize < expectedSize)
                    {
                        vlmData.setSizeState(Size.TOO_SMALL);
                    }
                    else
                    {
                        Size sizeState = Size.TOO_LARGE;

                        final long toleratedSize = expectedSize + extentSizes.get(info.volumeGroup) * TOLERANCE_FACTOR;
                        if (actualSize < toleratedSize)
                        {
                            sizeState = Size.TOO_LARGE_WITHIN_TOLERANCE;
                        }
                        vlmData.setSizeState(sizeState);
                    }
                }
                else
                {
                    vlmData.setSizeState(Size.AS_EXPECTED);
                }
            }
        }
    }

    private String getFullQualifiedIdentifier(SpdkData<?> vlmDataRef)
    {
        return vlmDataRef.getVolumeGroup() +
            File.separator +
            asIdentifierRaw(vlmDataRef);
    }

    @SuppressWarnings("unchecked")
    protected String asIdentifierRaw(SpdkData<?> vlmData)
    {
        String identifier;
        if (vlmData.getVolume() instanceof Volume)
        {
            identifier = asLvIdentifier((SpdkData<Resource>) vlmData);
        }
        else
        {
            identifier = asSnapLvIdentifier((SpdkData<Snapshot>) vlmData);
        }
        return identifier;
    }

    protected void updateInfo(SpdkData<?> vlmData, LvsInfo info)
        throws DatabaseException, AccessDeniedException, StorageException
    {
        vlmData.setIdentifier(asIdentifierRaw(vlmData));
        if (info == null)
        {
            vlmData.setExists(false);
            vlmData.setVolumeGroup(extractVolumeGroup(vlmData));
            vlmData.setDevicePath(null);
            // vlmData.setIdentifier(null);
            vlmData.setAllocatedSize(-1);
            // vlmData.setUsableSize(-1);
        }
        else
        {
            vlmData.setExists(true);
            vlmData.setVolumeGroup(info.volumeGroup);
            vlmData.setDevicePath(null); // if devicePath != null DevHandler will ask udev to get symlinks
            // pointing to that devPath which will fail
            vlmData.setSpdkPath(info.path);
            vlmData.setIdentifier(info.identifier);

            long size = SizeConv.convert(info.size, SizeUnit.UNIT_B, SizeUnit.UNIT_KiB);
            vlmData.setAllocatedSize(size);
            vlmData.setUsableSize(size);
        }
    }

    protected String extractVolumeGroup(SpdkData<?> vlmData)
    {
        return getVolumeGroup(vlmData.getStorPool());
    }

    @SuppressWarnings("unused")
    protected void updateSnapshotStates(Collection<SnapshotVolume> snapshots)
        throws AccessDeniedException, DatabaseException
    {
        // no-op
    }

    @Override
    protected void createLvImpl(SpdkData<Resource> vlmData)
        throws StorageException, AccessDeniedException
    {
        spdkCommands.createFat(
            vlmData.getVolumeGroup(),
            asLvIdentifier(vlmData),
            vlmData.getExpectedSize(),
            "--type=" + getLvCreateType(vlmData)
        );
    }

    protected String getLvCreateType(SpdkData<Resource> vlmDataRef)
    {
        String type;
        try
        {
            Volume vlm = (Volume) vlmDataRef.getVolume();
            ResourceDefinition rscDfn = vlm.getResourceDefinition();
            ResourceGroup rscGrp = rscDfn.getResourceGroup();
            VolumeDefinition vlmDfn = vlm.getVolumeDefinition();
            PriorityProps prioProps = new PriorityProps(
                vlm.getProps(storDriverAccCtx),
                vlmDfn.getProps(storDriverAccCtx),
                rscGrp.getVolumeGroupProps(storDriverAccCtx, vlmDfn.getVolumeNumber()),
                rscDfn.getProps(storDriverAccCtx),
                rscGrp.getProps(storDriverAccCtx),
                vlmDataRef.getStorPool().getProps(storDriverAccCtx)
            );
            type = prioProps.getProp(
                ApiConsts.KEY_STOR_POOL_LVCREATE_TYPE,
                ApiConsts.NAMESPC_STORAGE_DRIVER,
                DFLT_LVCREATE_TYPE
            );
        }
        catch (AccessDeniedException | InvalidKeyException exc)
        {
            throw new ImplementationError(exc);
        }
        return type;
    }

    @Override
    protected void resizeLvImpl(SpdkData<Resource> vlmData)
        throws StorageException, AccessDeniedException
    {
        spdkCommands.resize(
            vlmData.getVolumeGroup(),
            asLvIdentifier(vlmData),
            vlmData.getExpectedSize()
        );
    }

    @Override
    protected void deleteLvImpl(SpdkData<Resource> vlmData, String oldSpdkId)
        throws StorageException, DatabaseException, AccessDeniedException
    {
        // just make sure to not colide with any other ongoing wipe-lv-name
        String newSpdkId = String.format(
            FORMAT_SPDK_ID_WIPE_IN_PROGRESS,
            asLvIdentifier(vlmData)
        );

        String devicePath = vlmData.getSpdkPath();
        // devicePath is the "current" devicePath. as we will rename it right now
        // we will have to adjust the devicePath
        int lastIndexOf = devicePath.lastIndexOf(oldSpdkId);
        devicePath = devicePath.substring(0, lastIndexOf) + newSpdkId;
        String volumeGroup = vlmData.getVolumeGroup();

        spdkCommands.rename(
            volumeGroup,
            oldSpdkId,
            newSpdkId
        );

        vlmData.setExists(false);

        // SPDK by default wipes a lvol bdev during deletion, unless this option was disabled in lvol store
        try
        {
            spdkCommands.delete(
                volumeGroup,
                newSpdkId
            );
        }
        catch (StorageException exc)
        {
            errorReporter.reportError(exc);
        }
    }

    @Override
    protected void deactivateLvImpl(SpdkData<Resource> vlmDataRef, String lvIdRef)
        throws StorageException, AccessDeniedException, DatabaseException
    {
        // noop, not supported
    }

    @Override
    protected Map<String, Long> getFreeSpacesImpl() throws StorageException
    {
        Map<String, Long> freeSizes;
        try
        {
            freeSizes = SpdkUtils.getVgFreeSize(spdkCommands, changedStoragePoolStrings);
            for (String storPool : changedStoragePoolStrings)
            {
                if (!freeSizes.containsKey(storPool))
                {
                    freeSizes.put(storPool, SIZE_OF_NOT_FOUND_STOR_POOL);
                }
            }
        }
        catch (AccessDeniedException exc)
        {
            throw new ImplementationError(exc);
        }
        return freeSizes;
    }

    @Override
    protected Map<String, LvsInfo> getInfoListImpl(
        List<SpdkData<Resource>> vlmDataList,
        List<SpdkData<Snapshot>> snapVlms
    )
        throws StorageException, AccessDeniedException
    {
        return SpdkUtils.getLvsInfo(
            spdkCommands,
            getAffectedVolumeGroups(vlmDataList, snapVlms)
        );
    }

    @Override
    public String getDevicePath(String storageName, String lvId)
    {
        return String.format(SPDK_FORMAT_DEV_PATH, storageName, lvId);
    }

    @Override
    protected String asLvIdentifier(ResourceName resourceName, String rscNameSuffix, VolumeNumber volumeNumber)
    {
        return String.format(
            FORMAT_RSC_TO_SPDK_ID,
            resourceName.displayValue,
            rscNameSuffix,
            volumeNumber.value
        );
    }

    @Override
    protected String asSnapLvIdentifierRaw(String rscNameRef, String rscNameSuffixRef, String snapNameRef, int vlmNrRef)
    {
        return String.format(
            FORMAT_SNAP_TO_SPDK_ID,
            rscNameRef,
            rscNameSuffixRef,
            vlmNrRef,
            snapNameRef
        );
    }

    @Override
    protected String getStorageName(StorPool storPoolRef)
    {
        return getVolumeGroup(storPoolRef);
    }

    protected String getVolumeGroup(StorPool storPool)
    {
        String volumeGroup;
        try
        {
            volumeGroup = DeviceLayerUtils.getNamespaceStorDriver(
                storPool.getProps(storDriverAccCtx)
            )
                .getProp(StorageConstants.CONFIG_LVM_VOLUME_GROUP_KEY).split("/")[0];
        }
        catch (InvalidKeyException | AccessDeniedException exc)
        {
            throw new ImplementationError(exc);
        }
        return volumeGroup;
    }

    @Override
    protected boolean updateDmStats()
    {
        return false; // no path to a block device in SPDK
    }

    @Override
    public SpaceInfo getSpaceInfo(StorPool storPool) throws StorageException, AccessDeniedException
    {
        String vg = getVolumeGroup(storPool);
        if (vg == null)
        {
            throw new StorageException("Unset volume group for " + storPool);
        }
        Long capacity = SpdkUtils.getVgTotalSize(
            spdkCommands,
            Collections.singleton(vg)
        ).get(vg);
        Long freespace = SpdkUtils.getVgFreeSize(
            spdkCommands,
            Collections.singleton(vg)
        ).get(vg);

        return new SpaceInfo(capacity, freespace);
    }

    @Override
    protected boolean waitForSnapshotDevice()
    {
        return true;
    }

    @Override
    public LocalNodePropsChangePojo checkConfig(StorPool storPool) throws StorageException, AccessDeniedException
    {
        Props props = DeviceLayerUtils.getNamespaceStorDriver(
            storPool.getProps(storDriverAccCtx)
        );
        SpdkConfigReader.checkVolumeGroupEntry(spdkCommands, props);
        SpdkConfigReader.checkToleranceFactor(props);
        return null;
    }

    @Override
    protected long getAllocatedSize(SpdkData<Resource> vlmDataRef) throws StorageException
    {
        try
        {
            return SpdkUtils.getBlockSizeByName(
                spdkCommands,
                vlmDataRef.getSpdkPath().split(SPDK_PATH_PREFIX)[1]
            );
        }
        catch (AccessDeniedException exc)
        {
            throw new ImplementationError(exc);
        }
    }

    private Set<String> getAffectedVolumeGroups(
        Collection<SpdkData<Resource>> vlmDataList,
        Collection<SpdkData<Snapshot>> snapVlms
    )
    {
        ArrayList<SpdkData<?>> combinedList = new ArrayList<>();
        combinedList.addAll(vlmDataList);
        combinedList.addAll(snapVlms);

        Set<String> volumeGroups = new HashSet<>();
        for (SpdkData<?> vlmData : combinedList)
        {
            String volumeGroup = vlmData.getVolumeGroup();
            if (volumeGroup == null)
            {
                volumeGroup = getVolumeGroup(vlmData.getStorPool());
                vlmData.setVolumeGroup(volumeGroup);
            }
            if (volumeGroup != null)
            {
                volumeGroups.add(volumeGroup);
            }
        }
        return volumeGroups;
    }

    @Override
    public void update(StorPool storPoolRef) throws AccessDeniedException, DatabaseException, StorageException
    {
        // TODO: we need to implement a check for pmem here. something like LvmProvider.update does
    }

    @Override
    protected void setDevicePath(SpdkData<Resource> vlmData, String devPath) throws DatabaseException
    {
        vlmData.setSpdkPath(devPath);
    }

    @Override
    protected void setAllocatedSize(SpdkData<Resource> vlmData, long size) throws DatabaseException
    {
        vlmData.setAllocatedSize(size);
    }

    @Override
    protected void setUsableSize(SpdkData<Resource> vlmData, long size) throws DatabaseException
    {
        vlmData.setUsableSize(size);
    }

    @Override
    protected void setExpectedUsableSize(SpdkData<Resource> vlmData, long size)
    {
        vlmData.setExepectedSize(size);
    }

    @Override
    protected String getStorageName(SpdkData<Resource> vlmDataRef) throws DatabaseException
    {
        return vlmDataRef.getVolumeGroup();
    }

    public SpdkCommands<?> getSpdkCommands()
    {
        return spdkCommands;
    }
}
